(*
  #****m* SecDec/general/src/subexp/formContourC.m
  #  NAME
  #    formContourC.m
  #
  #  USAGE
  #  is called by subandexpand*l*h*.m to complete the epsilon expansion, and write the C++ files f*.cc
  #  and g*.cc into the appropriate subdirectories
  # 
  #  USES 
  #  parts.m, ExpOpt.m, 
  #
  #  USED BY 
  #    
  #  subandexpand*l*h*.m
  #
  #  PURPOSE
  #  completes the epsilon expansion, does the subtraction and contour deformation and writes the 
  #  C++ files f*.cc and g*.cc in the appropriate subdirectories
  #    
  #  INPUTS
  #  from subandexpand*l*h*:
  #  n: number of propagators
  #  expu, expf: exponents of U and F^-1 respectively
  #  path, srcdir: where to load parts.m, ExpOpt.m from
  #  logi, lini, higheri: the number of logarithmic, linear and higher order poles respectively
  #
  #  originally from formindlist.m:
  #  integrandfunctionlist: contains the list of exponents of each variable, together with the number of functions
  #  with the identical exponent structure
  #  fstore[*,*],ustore[*,*],nstore[*,*],degen[*,*]: the numerator, and functions f and u from each subsector (after
  #  decomposition and permutation of variables to exploit symmetries of the problem), together with any information on
  #  degeneracies of these functions (eg if subsector A == subsector B upto a permutations of variables)
  #   
  #  originally from symbsub.m:
  #  epspower[*]:the power of epsilon as a prefactor in piece * of the subtraction
  #  numcoeff[*]:the O(1) prefactor of the piece * of the subtraction
  #  set[*]: if {x,a} were an element of set[*], this indicates that the piece * of the subtraction is to 
  #   be differentiated 'a' times wrt x, and x is then to be set to zero
  #  exponents[*,**]: the exponent of variable z[**] in piece * of the subtraction
  #
  #  GLOBAL VARIABLES
  #  functioncounter[*]:counts the number of functions f and g for each order in epsilon
  #
  #  RESULT
  #  C++ functions f*.cc and g*.cc are written in the appropriate subdirectory corresponding to the given graph, 
  #  pole structure, order in epsilon, and, when IBP is used, the number of independent variables in f*.cc
  #
  #  SEE ALSO
  #  subandexpand*l*h*.m, formindlist.m, symbsub.m, parts.m, ExpOpt.m
  #   
  #****
 *)
Get[StringJoin[path,"src/deco/sqrtsimp.m"]];
mygamma[a_+b_ eps]:=Module[{},
If[And[NumberQ[a],a<=0],1/(a+b eps) mygamma[a+1+b eps],Gamma[a+b eps]]];

mygamma[b_ eps]:=1/(b eps) Gamma[1+b eps];
mygamma[eps]:=1/eps Gamma[1+eps];
mygamma[a_+eps]:=If[And[NumberQ[a],a<=0],1/(a+eps)*
mygamma[a+1+eps],Gamma[a+eps]];

comptest[ca_,cb_]:=Module[{},
			  If[MatchQ[ca[[1]],cb[[1]]],True,
			     And[ca[[1]]>-1,cb[[1]]>-1]]];
sametest[sa_,sb_]:=Module[{tlist},
			  tlist=MapThread[comptest,{sa[[1]],sb[[1]]}];
			  tlist=Flatten[tlist/.True->{}];
			  If[MatchQ[tlist,{}],True,False]];

numberab[ni_]:=Block[{exps},
		      exps=integrandfunctionlist[[ni,1]];
		      Do[a[numberi]=exps[[numberi,1]];b[numberi]=exps[[numberi,2]],{numberi,n-1}];
		      ];
(*
 funset[fsj_,fsk_,fsset_]:=
 (Fold[der,(fstore[fsj,fsk]^(-expf)ustore[fsj,fsk]^(expu)nstore[fsj,fsk]),fsset]); 
 *)

funset[fsj_,fsk_,fsset_]:=
(Fold[der,(store[fsj,fsk])/.Gamma->mygamma,fsset]);

der[derfun_,derset_]:=(D[derfun,derset])/.derset[[1]]->0;

epsexpand[exprtoex_,ordertoex_]:=Table[((D[exprtoex,{eps,epsdif}])/.eps->0)/epsdif!
				       ,
				       {epsdif,0,ordertoex}
				       ];

epsmulti[l1_,l2_,ordtom_]:=Table[Table[l1[[i]]l2[[j-i+1]],{i,j}],{j,ordtom+1}];


populateintlists[poplist_,poppow_]:=Block[{temppow},
					  temppow=poppow;
					  {integrand[temppow]={integrand[temppow],#},temppow++}&/@poplist;
					  ];

genTau[gtf_]:=Block[{i,taupre,taufun},
		    taupre[i_]:=xvals[0] lrs[i-1] z[i] (1-z[i]); 
		    taufun[i_]:=D[gtf,z[i]];
		    Do[eTau[i]=taupre[i]*taufun[i],{i,n-1}];
		    Return[eTau]
		    ];

Deformation[deff_]:=Block[{OV,NV,JACMAT,detjac,defreps},
			  TAU=genTau[deff];
			  OV=Table[z[i],{i,n-1}];
			  NV=Table[nz[i]=z[i]-I TAU[i],{i,n-1}];
			  JACMAT=Table[D[NV[[i]],OV[[j]]],{i,n-1},{j,n-1}];
			  detjac=Det[JACMAT];
			  defreps=Table[z[i]->z[i]-I TAU[i],{i,n-1}];
			  Return[{detjac,defreps,TAU}]
			  ];

argexp[expr_]:=Block[{temp,dxarg=0,i,arg},
		     temp = modexp[expr];
		     arg = temp[[3]]/temp[[2]]/xvals[0];
		     (*Do[dxarg += Abs[D[arg,z[i]]],{i,n-1}];*)
		     Return[arg]
		     ];

modexp[expr_]:=Block[{temp,smexpand,xv0,repa,impa,a,b},
		     xv0/:Power[xv0,a_]:=0/;OddQ[a];
		     xv0/:Power[xv0,a_]:=xvals[0]^a/;EvenQ[a];
		     smexpand=Collect[expr,xvals[0]];
		     temp=smexpand/.xvals[0]->xv0;
		     repa=temp/.xv0->0;
		     impa=(smexpand-repa)/.Complex[a_,b_]->b;
		     Return[{repa^2+impa^2,repa,impa}]
		     ];

ratexp[expr_]:=Block[{a,b},
		     Return[(Coefficient[expr,xvals[0],3]/.Complex[a_,b_]->b)/(Coefficient[expr,xvals[0],1]/.Complex[a_,b_]->b)]
		     ];

expon[exa_,exb_]:=integrandfunctionlist[[exa,1,exb,1]]+integrandfunctionlist[[exa,1,exb,2]] eps;

(***************** BEGIN FUNCTIONS for forming integrand and writing files **************************)

optimlambdafunC[gepsord_,functioncounter_,gcondeff_,gcondef3_]:=
  Block[{fexpanded,ratiotest,modtest,a,signtest,tautest,argtest,ratioopt,modopt,signopt,argopt,tauopt,goutfile,rfuncname,mfuncname,sfuncname,tfuncname,afuncname,try,tt,he1},
	fexpanded=Expand[gcondeff,xvals[0]];
	(*debugfilename=StringJoin[direy,"/g",ToString[functioncounter[-2]],".m"];
	 Save[debugfilename,fexpanded];*)
	
	ratiotest=ratexp[fexpanded]/.{xvals[0]->lambda};
	{modtest,a,signtest}=modexp[fexpanded]/.{xvals[0]->lambda}; 
	tautest=gcondef3;
	
	(*Share[];*) (*comment out if forming of integrand should be faster*)
	
	try=OptimizeExpression[ratiotest, OptimizationSymbol -> y];
	tt =try/.{OptimizedExpression->Hold};
	he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	ratioopt=ReleaseHold[he1];
	
	try=OptimizeExpression[modtest, OptimizationSymbol -> y]; 
	tt =try/.{OptimizedExpression->Hold};
	he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	modopt=ReleaseHold[he1];
	
	try=OptimizeExpression[signtest, OptimizationSymbol -> y];
	tt =try/.{OptimizedExpression->Hold};
	he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};					
	signopt=ReleaseHold[he1];
	If[oscillatory==1,
	   argtest=argexp[fexpanded]/.xvals[0]->lambda;
	   try=OptimizeExpression[argtest, OptimizationSymbol -> y];
	   tt =try/.{OptimizedExpression->Hold};
	   he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};	
	   argopt=ReleaseHold[he1];
	   ];
	
	Do[
	   try=OptimizeExpression[tautest[i]/.{xvals[0]->1,lrs[i-1]->1}, OptimizationSymbol -> y];
	   tt =try/.{OptimizedExpression->Hold};
	   he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	   tauopt[i]=ReleaseHold[he1];
	   ,{i,n-1}
	   ];
	
	direy=StringJoin[direyt,"/epstothe",ToString[gepsord]];
	If[FileNames[direy]=={},CreateDirectory[direy]];
	
	goutfile=StringJoin[direy,"/g",ToString[functioncounter],".cc"];
	
	rfuncname=StringJoin["P",polestring,"r",ToString[functioncounter]];
	mfuncname=StringJoin["P",polestring,"m",ToString[functioncounter]];
	sfuncname=StringJoin["P",polestring,"s",ToString[functioncounter]];
	afuncname=StringJoin["P",polestring,"a",ToString[functioncounter]];
	tfuncnamepre=StringJoin["P",polestring,"t",ToString[functioncounter],"x"];
	
	writeoptC[ratioopt,varletter,rfuncname,goutfile,1];
	writeoptC[modopt,varletter,mfuncname,goutfile,2];
	writeoptC[signopt,varletter,sfuncname,goutfile,2];
	If[oscillatory==1, writeoptC[argopt,varletter,afuncname,goutfile,2];];
	Do[
	   writeoptC[tauopt[i],varletter,tfuncnamepre<>ToString[i],goutfile,2];
	   ,{i,n-1}];
	Clear[try,tt,he1]
	];

formnumericalintegrandC[explab_,epsordreq_,0]:=
  Block[{condef,condeff,condef3,counter,epsord,prods,mufunct,expnumfact,exorder,expnfeps},
	numberab[explab];
	Clear[integrand];
	Do[integrand[inti]={},{inti,-minpole,epsordreq-ordcheck}];
	Do[
	   (*****BEGIN deformation of contour***********)
	   condef=Deformation[store[explab,functionnumber]];
	   store[explab,functionnumber]=store[explab,functionnumber]/.condef[[2]];
	   prods=Product[(1-I TAU[i]/z[i])^expon[explab,i],{i,n-1}];
	   (*
	    nstore[explab,functionnumber]=nstore[explab,functionnumber]*condef[[1]]*prods;
	    *)
	   condeff=store[explab,functionnumber];
	   condef3=condef[[3]];
	   Clear[condef];
	   (*****END deformation of contour***********)
	   (*****BEGIN subtraction******)
	   Do[If[
		 epspower[mu]<=epsordreq-ordcheck
		 ,
		 mufunct=Table[
			       mysimp[funset[explab,functionnumber,set[mu]],zlist]
			       ,
			       {functionnumber,integrandfunctionlist[[explab,2]]}
			       ];
		 mufunct=mufunct//.List[a__,0,b__]->List[a,b];
		 expnumfact=Product[z[expi]^exponents[mu,expi],{expi,n}]numcoeff[mu];
		 exorder=epsordreq-epspower[mu]-ordcheck;
		 expnfeps=epsexpand[expnumfact,exorder];
		 mufunct=Function[xx,epsexpand[xx,exorder]]/@mufunct;
		 mufunct=Function[xx,epsmulti[expnfeps,xx,exorder]]/@mufunct;
		 Function[xx,populateintlists[xx,epspower[mu]]]/@mufunct;	
		 ];
	      ,{mu,sizemu}];
	   Clear[mufunct,expnfeps];
	   (*****END subtraction******)
	   (*****BEGIN write f*.c and g*.c functions*******)
	   listel=1;
	   Do[integrands[epsord]=Plus@@Flatten[integrand[epsord]],{epsord,-minpole,epsordreq-ordcheck}];
	   Do[If[MatchQ[integrand[epsord],{}]==False
		 ,
		 functiontooptimize=expandedimp[[listel]]/.{integimp->integrand,constimp->constepspre};
		 listel++;
		 If[MatchQ[functiontooptimize,0]==False,
		    counter=++functioncounter[epsord];
		    direy=StringJoin[direyt,"/epstothe",ToString[epsord]];
		    If[FileNames[direy]=={},CreateDirectory[direy]];
		    foutfile=StringJoin[direy,"/f",ToString[counter],".cc"];
		    moutfile=StringJoin[direy,"/f",ToString[counter],".m"];
		    ffuncname=StringJoin["P",polestring,"f",ToString[counter]];
		    
		    (******BEGIN Change of variables for endpoints*******)
		    If[endpointflag==1,
		       funcpart1=(functiontooptimize/.(z[a_]->z[a]/2))/2;
		       try=OptimizeExpression[funcpart1, OptimizationSymbol -> y];
		       tt =try/.{OptimizedExpression->Hold};
		       he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
		       expr=ReleaseHold[he1];
		       writeoptC[expr,varletter,ffuncname,foutfile];
		       optimlambdafunC[epsord,counter,condeff,condef3];
		       (*Save[moutfile,funcpart1];*)
		       Clear[funcpart1];
		       Share[];
		       
		       counter=++functioncounter[epsord];
		       foutfile=StringJoin[direy,"/f",ToString[counter],".cc"];
		       ffuncname=StringJoin["P",polestring,"f",ToString[counter]];
		       funcpart2=(functiontooptimize/.(z[a_]->1-z[a]/2))/2;
		       try=OptimizeExpression[funcpart2, OptimizationSymbol -> y];
		       tt =try/.{OptimizedExpression->Hold};
		       he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
		       expr=ReleaseHold[he1];
		       writeoptC[expr,varletter,ffuncname,foutfile];
		       optimlambdafunC[epsord,counter,condeff,condef3];
		       (*functiontooptimize=Plus@@{funcpart1,funcpart2}/2;*)
		       (*Save[moutfile,funcpart2];*)
		       Clear[funcpart2]
		       ,
		       try=OptimizeExpression[functiontooptimize, OptimizationSymbol -> y];
		       tt =try/.{OptimizedExpression->Hold};
		       he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
		       expr=ReleaseHold[he1];
		       writeoptC[expr,varletter,ffuncname,foutfile];
		       (* moutfile introduced for debugging purposes only: Save[moutfile,functiontooptimize]; *)
		       optimlambdafunC[epsord,counter,condeff,condef3];
		       ];
		    (******END Change of variables for endpoints*******)
		    ]
		 Clear[functiontooptimize,try,tt,he1,expr];
		 ]
	      ,
	      {epsord,-minpole+ordcheck,epsordreq}
	      ]; 
	   (*****END write f*.c and g*.c functions*******)
	   ]
	Clear[a,b];
	];

If[partsflag==1,
   Get[StringJoin[path,"src/subexp/","parts.m"]];
   ];

formnumericalintegrandC[explab_,epsordreq_,1]:=
  Block[{condef,condeff,condef3,prods,expt,exptifl,expdif,mufunctpre,mui,mu,inpa,varsset,mufunct,functionnumber},
	numberab[explab];
	Clear[integrands,memcount];
	Do[Do[integrands[epsfort,epspole]={};memcount[epsfort,epspole]=0,{epspole,epsfort,epsordreq}],{epsfort,-minpole,Min[epsordreq,0]}];
	Do[If[MatchQ[degen[explab,functionnumber],0]==False
	      ,
	      (*****BEGIN deformation of contour***********)
	      condef=Deformation[fstore[explab,functionnumber]];
	      (#[explab,functionnumber]=#[explab,functionnumber]/.condef[[2]])&/@{fstore,ustore,nstore};
	      prods=Product[(1-I TAU[i]/z[i])^expon[explab,i],{i,n-1}];
	      nstore[explab,functionnumber]=nstore[explab,functionnumber]*condef[[1]]*prods;
	      condeff=fstore[explab,functionnumber];
	      condef3=condef[[3]];
	      Clear[condef];
	      (*****END deformation of contour***********)
	      (*****BEGIN subtraction******)
	      Do[
		 If[epspower[mu]<=epsordreq,
		    mufunctpre=1;
		    varsset={};
		    Do[
		       expt=exponents[mu,mui]/.eps->0;
		       exptifl=integrandfunctionlist[[explab,1,mui,1]];
		       expdif=expt-exptifl;
		       If[
			  expt <0
			  ,
			  mufunctpre=mufunctpre*z[mui]^expdif;
			  inpa=integrandfunctionlist[[explab,1,mui,2]];
			  varsset=Append[varsset,{z[mui],inpa,exptifl}];
			  ,
			  mufunctpre=mufunctpre*z[mui]^(exponents[mu,mui])
			  ];
		       ,{mui,n-1}
		       ];
		    mufunct=mufunctpre*degen[explab,functionnumber]*funset[explab,functionnumber,set[mu]];
		    ftoC[mufunct,mu,epsordreq,numcoeff[mu],varsset];
		    (*Share[];*)
		    Clear[mufunct,varsset,expt];
		    ]
		 ,{mu,sizemu}];
	      Clear[condeff,condef3];
	      (*****END subtraction******)
	      ],{functionnumber,integrandfunctionlist[[explab,2]]}];
	Clear[a,b];
	];

WRITEFUNCTION[wmu_,wepspole_]:=
  Block[{functiontooptimize,try,tt,he1,expr,funcpart1,funcpart2,foutfile,direy,ffuncname,counter},
	functiontooptimize=Plus@@(Flatten[integrands[epspower[wmu],wepspole]]);
	memcount[epspower[wmu],wepspole]=0;
	integrands[epspower[wmu],wepspole]={};
	direy=StringJoin[direyt,"/epstothe",ToString[wepspole]];
	If[FileNames[direy]=={},CreateDirectory[direy]];
	If[
	   MatchQ[functiontooptimize,0]==False
	   ,
	   counter=++functioncounter[wepspole];
	   (*Print["lock ",counter," epspole ",wepspole];*)
	   foutfile=StringJoin[direy,"/f",ToString[counter],".cc"];
	   (*moutfile=StringJoin[direy,"/f",ToString[counter],".m"];*)
	   ffuncname=StringJoin["P",polestring,"f",ToString[counter]];
	   
	   functiontooptimize=functiontooptimize/.{Complex[aC_,bC_]->aC+bC MYI,xvals[0]->lambda};
	   (******BEGIN Change of variables for endpoints*******)
	   If[endpointflag==1,
	      funcpart1=(functiontooptimize/.(z[a_]->z[a]/2))/2;
	      try=OptimizeExpression[funcpart1, OptimizationSymbol -> y];
	      tt =try/.{OptimizedExpression->Hold};
	      he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	      expr=ReleaseHold[he1];
	      writeoptC[expr,varletter,ffuncname,foutfile];
	      optimlambdafunC[epsord,counter,condeff,condef3];
	      Clear[funcpart1];
	      Share[];
	      
	      counter=++functioncounter[wepspole];
	      foutfile=StringJoin[direy,"/f",ToString[counter],".cc"];
	      ffuncname=StringJoin["P",polestring,"f",ToString[counter]];
	      funcpart2=(functiontooptimize/.(z[a_]->1-z[a]/2))/2;
	      try=OptimizeExpression[funcpart2, OptimizationSymbol -> y];
	      tt =try/.{OptimizedExpression->Hold};
	      he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	      expr=ReleaseHold[he1];
	      writeoptC[expr,varletter,ffuncname,foutfile];
	      optimlambdafunC[epsord,counter,condeff,condef3];
	      (*functiontooptimize=Plus@@{funcpart1,funcpart2}/2;*)
	      Clear[funcpart2]
	      ,
	      try=OptimizeExpression[functiontooptimize, OptimizationSymbol -> y];
	      tt =try/.{OptimizedExpression->Hold};
	      he1=tt/.{Block->MyBlock,CompoundExpression->List,Set->Rule};
	      expr=ReleaseHold[he1];
	      writeoptC[expr,varletter,ffuncname,foutfile];
	      (* moutfile introduced for debugging purposes only: Save[moutfile,functiontooptimize]; *)
	      optimlambdafunC[epsord,counter,condeff,condef3];
	      ];
	   (******END Change of variables for endpoints*******)
	   Clear[functiontooptimize,try,tt,he1,expr];
	   ]
	];

EXPLICITIZE[Fmf_,emu_,eepsreq_,Tepsord_,Epsexpandedfunct_,postDpart_]:=
  Block[{pDa,pDi,postDmf,TIMESmf,thisepsorder,thisepsexpandedfunct,explicitmf,explicitTIMES,epspole},
	Clear[explicitmf,explicitTIMES];
	explicitmf[emfa__]:=0;
	lenemfa[emfa_]:=0;
	thisepsorder=Tepsord;						       
	If[ListQ[Fmf],
	   postDmf=Fmf[[1]];
	   TIMESmf=Times@@Fmf[[2]];
	   Do[
	      explicitmf[emfa]=postepsD[postDmf,{eps,emfa}];
	      explicitmf[emfa]=explicitmf[emfa]/.eps->0;
	      If[MatchQ[Head[explicitmf[emfa]],List]
		 ,
		 explicitmf[emfa]=Flatten[explicitmf[emfa]];
		 lenemfa[emfa]=Length[explicitmf[emfa]];
		 Do[explicitmf[emfa,emfb]=((explicitmf[emfa][[emfb]])/.postD->D)/emfa!
		    ,{emfb,lenemfa[emfa]}]
		 ,explicitmf[emfa,1]=((explicitmf[emfa])/.postD->D)/emfa!;
		 lenemfa[emfa]=1;
		 ];
	      explicitmf[emfa]=0;
	      explicitTIMES[emfa]=((D[TIMESmf,{eps,emfa}])/.eps->0)/emfa!;
	      ,{emfa,0,thisepsorder}]
	   ,TIMESmf=Fmf;
	   Do[
	      explicitTIMES[emfa]=((D[TIMESmf,{eps,emfa}])/.eps->0)/emfa!
	      ,{emfa,0,thisepsorder}];
	   explicitmf[0,1]=1;lenemfa[0]=1;
	   ];
	postDpart[pDa_]:=Table[postDpart[pDa,pDi],{pDi,lenemfa[pDa]}];
	thisepsexpandedfunct=Flatten/@Epsexpandedfunct;
	currentepsorder=epspower[emu];
	Share[];
	(*Writes f- and g-functions if memory cutoff is reached or crossed:*)
	INTEGRANDIZE[#,emu,postDpart,explicitmf,explicitTIMES]&/@thisepsexpandedfunct;
	Clear[explicitmf,explicitTIMES,postDmf,TIMESmf];
	Share[];
	Do[
	   If[And[MatchQ[integrands[epspower[emu],epspole],{}]==False,memcount[epspower[emu],epspole]>0],
	      (*Print["Write function, eps order = ",epspole];*)
	      WRITEFUNCTION[emu,epspole];
	      ]
	   ,{epspole,Max[-minpole,epspower[emu]],eepsreq}
	   ];
	];

INTEGRANDIZE[intlist_,imu_,PostDpart_,iexplicitmf_,iexplicitTIMES_]:=
  Block[{nextelement,nextmemcount},
	If[ListQ[intlist],INTEGRANDIZE[#,imu,PostDpart,iexplicitmf,iexplicitTIMES]&/@intlist;currentepsorder++,
	   nextelement=intlist/.{TIMESpart[Tpa_]->iexplicitTIMES[Tpa]};
	   nextelement=nextelement/.PostDpart->iexplicitmf;
	   nextmemcount=ByteCount[nextelement];
	   If[And[MatchQ[integrands[epspower[imu],currentepsorder],{}]==False,memcount[epspower[imu],currentepsorder]+nextmemcount>MEMCUTOFF],
	      Print["Threshold reached, eps order = ",currentepsorder,", memcount = ",memcount[epspower[imu],currentepsorder]];
	      WRITEFUNCTION[imu,currentepsorder];
	      ];
	   integrands[epspower[imu],currentepsorder]={integrands[epspower[imu],currentepsorder],nextelement};
	   memcount[epspower[imu],currentepsorder]=memcount[epspower[imu],currentepsorder]+nextmemcount;
	   ];
	Clear[nextelement,nextmemcount];
	];
	
ftoC[ffort_,fmu_,fepsreq_,nfort_,varsfort_]:=
  Block[{mf,thisepsorder,epsexpandedfunct,postDpart,postDlist},
	mf=Flatten[intparts[ffort,varsfort,precisionrequired-epspower[fmu]]];
	mf=mf//.{A___,0,B___}->{A,B};
	If[MatchQ[mf,{}]==False,
	   mf=((#/.ipsum->ipexplicitsum)*nfort)&/@mf;
	   mf=mf/.postD->postDtemp;
	   mf=mf/.Times[TB___,postDtemp[pDf__],TA___]->{postDtemp[pDf],TIMES[TB,TA]};
	   mf=mf/.postDtemp->postD;
	   thisepsorder=precisionrequired-epspower[fmu];
	   Clear[postDpart];
	   
	   postDlist=Table[postDpart[pDpi],{pDpi,0,thisepsorder}];
	   TIMESlist=Table[TIMESpart[tpi],{tpi,0,thisepsorder}];
	   epsexpandedfunct=epsmulti[postDlist,TIMESlist,thisepsorder];
	   EXPLICITIZE[#,fmu,fepsreq,thisepsorder,epsexpandedfunct,postDpart]&/@mf;
	   Share[];
	   ]
	];
(***************** END FUNCTIONS for forming integrand and writing files ****************************)
direy=StringJoin[outp,"/",polestring]; (*directory for the output files*)
direycount=0;
direyt=direy;
While[
      FileNames[direyt]=!={}
      ,
      direycount++;
      direyt=StringJoin[direy,ToString[direycount]]
      ];
If[direycount>0,RenameDirectory[direy,direyt]];(*puts old results into another directory, ~diagramname#, where the
						largest # relates to the most recent folder*)
CreateDirectory[direy]; (*Creates the directory to save the files to. Most recent directory is ~diagramname*)
direyt=direy;
minpole = logi+lini+higheri;
varletter="w";
MyBlock[listvar_,listabbr_]:={listvar,listabbr};
Clear[x];
Do[
   z[changezi]=ToExpression[StringJoin["x",ToString[changezi]]]
   ,
     {changezi,n}
   ];
zlist=Table[z[i],{i,n}];


diffexpos=Split[integrandfunctionlist,sametest];

red[a_]:={a[[1,1]],Length[a]};
diff=red/@diffexpos;
runsneeded=Length[diff];
rmaxim=0;

constpre=constpre//.Gamma->mygamma;

If[FreeQ[constpre,eps],ordcheck=0;constepspre[0]=constpre;constepspre[AA_]:=0,
 ordcheck=Series[constpre,{eps,0,-10}][[4]];
 constpreser=Series[constpre,{eps,0,minpole+precisionrequired}];
 Do[
  constepspre[cep]=SeriesCoefficient[constpreser,cep]
 ,
  {cep,ordcheck,minpole+precisionrequired}
 ]
];
constimptab=Table[constimp[cep],{cep,ordcheck,minpole+precisionrequired}];
integimptab=Table[integimp[imt],{imt,-minpole,precisionrequired-ordcheck}];
expandedimp=Plus@@#&/@epsmulti[constimptab,integimptab,precisionrequired+minpole-ordcheck];




ms[i_]:=em[i-1]; (*defines the remapping of ms*)
ssp[i_]:=esx[i-1];(*and ssp*)
xlambda=xvals[0];
(****************** Strings for C++ functions **********************)
Cstring0=";";
Cstring1="#include \"intfile.hh\"\n";
Cstring3=Cstring1<>"\ndouble ";
Cstring1=Cstring1<>"\ndcmplx ";
Cstring2="(const double x[], double es[], double esx[], double em[], double lambda, double lrs[]) {\n";
Cstring4=StringJoin@@("double x"<>#<>"=x["<>#<>"];\n"&/@Table[ToString[i],{i,0,n-2}]);
Cstring4a[co_,costr_:"dcmplx"]:=costr<>" "<>varletter<>"["<>ToString[co]<>"];\n";
Cstring5[costr_:"dcmplx"]:=costr<>" FOUT;\n";
Cstring7="dcmplx MYI(0.,1.);\n";
Cstring6="return ";
(******************** END Strings ***********************************)

(******** START formnumerical integrand and write functions *********)
createoptimizedC[cofrmin_,cofrmax_]:=Block[{tiepspow,numintegdo,infostream},
			Print["Memory in use before optimizing C++ functions = ",MemoryInUse[]];
			Do[functioncounter[tiepspow]=0,{tiepspow,-minpole+ordcheck,precisionrequired}];
			Do[
			   formnumericalintegrandC[numintegdo,precisionrequired,partsflag];
			   If[
			      Or[numintegdo<10,Mod[numintegdo,10]==0,numintegdo==Length[integrandfunctionlist]]
			      ,
			      Print["numericalintegrand ", numintegdo, " evaluated, Memory in use = ",MemoryInUse[]]
			      ];
			   ,{numintegdo,cofrmin,cofrmax}
			   ];
			While[functioncounter[-minpole]==0,minpole--];
			infostream = OpenWrite[StringJoin[direyt,"/infofile"]];
			Do[Write[infostream, StringJoin[ToString[tiepspow],"functions = ",ToString[functioncounter[tiepspow]]]]
			   ,{tiepspow,-minpole+ordcheck,precisionrequired}];
			Close[infostream];
			];
(********** END formnumerical integrand and write functions *********)

Print["Producing C++ functions"];

forttime=Timing[Do[symbolicsubtraction[diff[[rni,1]]];
		   rminim=rmaxim+1;
		   rmaxim=rmaxim+diff[[rni,2]];
		   createoptimizedC[rminim,rmaxim]
		   ,{rni,runsneeded}];
		][[1]];

Print["C++ functions produced, time taken = ",forttime," seconds"];
(*Print[??Global`*];
 Print["BC: ",ByteCount[goutfile]];
 Print["BC: ",ByteCount[minpole]];
 Print["BC: ",ByteCount[memcount[0,0]]];
 *)
